__author__ = "Lukas Leufen"
__date__ = "2022-07-05"


from typing import Union, List, Dict

from . import join, toar_data_v2

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
import pandas as pd


class EmptyQueryResult(Exception):
    """Exception that get raised if a query to JOIN returns empty results."""

    pass


def create_url(base: str, service: str, param_id: Union[str, int, None] = None,
               **kwargs: Union[str, int, float, None]) -> str:
    """
    Create a request url with given base url, service type and arbitrarily many additional keyword arguments.

    :param base: basic url of the rest service
    :param service: service type, e.g. series, stats
    :param param_id: id for a distinct service, is added between ending / of service and ? of kwargs
    :param kwargs: keyword pairs for optional request specifications, e.g. 'statistics=maximum'

    :return: combined url as string
    """
    url = f"{base}"
    if not url.endswith("/"):
        url += "/"
    if service is not None:
        url = f"{url}{service}"
    if not url.endswith("/"):
        url += "/"
    if param_id is not None:
        url = f"{url}{param_id}"
    if len(kwargs) > 0:
        url = f"{url}?{'&'.join(f'{k}={v}' for k, v in kwargs.items() if v is not None)}"
    return url


def get_data(opts: Dict, headers: Dict, as_json: bool = True) -> Union[Dict, List, str]:
    """
    Download join data using requests framework.

    Data is returned as json like structure. Depending on the response structure, this can lead to a list or dictionary.

    :param opts: options to create the request url
    :param headers: additional headers information like authorization, can be empty
    :param as_json: extract response as json if true (default True)

    :return: requested data (either as list or dictionary)
    """
    url = create_url(**opts)
    try:
        response = retries_session().get(url, headers=headers, timeout=(5, None))  # timeout=(open, read)
        if response.status_code == 200:
            return response.json() if as_json is True else response.text
        else:
            raise EmptyQueryResult(f"There was an error (STATUS {response.status_code}) for request {url}")
    except requests.exceptions.RetryError as e:
        raise EmptyQueryResult(f"There was an RetryError for request {url}: {e}")


def retries_session(max_retries=3):
    retry_strategy = Retry(total=max_retries,
                           backoff_factor=0.1,
                           status_forcelist=[429, 500, 502, 503, 504],
                           method_whitelist=["HEAD", "GET", "OPTIONS"])
    adapter = HTTPAdapter(max_retries=retry_strategy)
    http = requests.Session()
    http.mount("https://", adapter)
    http.mount("http://", adapter)
    return http


def download_toar(station, toar_stats, sampling, data_origin):

    try:
        # load data from toar-data (v2)
        df_toar, meta_toar = toar_data_v2.download_toar(station, toar_stats, sampling=sampling, data_origin=data_origin)
    except (AttributeError, EmptyQueryResult, KeyError, requests.ConnectionError, ValueError, IndexError):
        df_toar, meta_toar = None, None

    try:
        # load join data (toar-data v1)
        df_join, meta_join = join.download_join(station_name=station, stat_var=toar_stats, sampling=sampling,
                                                data_origin=data_origin)
    except (AttributeError, EmptyQueryResult, KeyError, requests.ConnectionError, ValueError, IndexError):
        df_join, meta_join = None, None

    # merge both data sources with priority on toar-data v2
    if df_toar is not None and df_join is not None:
        df_merged = merge_toar_join(df_toar, df_join, sampling)
        meta_merged = meta_toar
    else:
        df_merged = df_toar if df_toar is not None else df_join
        meta_merged = meta_toar if df_toar is not None else meta_join
    return df_merged, meta_merged


def merge_toar_join(df_toar, df_join, sampling):
    start_date = min([df_toar.index.min(), df_join.index.min()])
    end_date = max([df_toar.index.max(), df_join.index.max()])
    freq = {"hourly": "1H", "daily": "1d"}.get(sampling)
    full_time = pd.date_range(start_date, end_date, freq=freq)
    full_data = df_toar.reindex(full_time)
    full_data.update(df_join, overwrite=False)
    return full_data


def correct_stat_name(stat: str) -> str:
    """
    Map given statistic name to new namespace defined by mapping dict.

    Return given name stat if not element of mapping namespace.

    :param stat: namespace from JOIN server

    :return: stat mapped to local namespace
    """
    mapping = {'average_values': 'mean', 'maximum': 'max', 'minimum': 'min'}
    return mapping.get(stat, stat)
